@@ -2,7 +2,7 @@
 /*
     no-ip.com dynamic IP update client for Linux
 
-   Copyright 2000-2004 Free Software Foundation, Inc.
+   Copyright 2000-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -87,6 +87,31 @@
 
 	+	January 2004 (johna)	version 2.1.1
 	+	added -u, -p and -x options for LR101 project
+
+	+	April 2004 (johna	version 2.1.2
+	+	removed -Y in make install rule
+
+	+	August 2005 (johna)	version 2.1.3
+	+	added shm dump code for debugging broken libraries
+	+	added -z flag to invoke shm dump code
+
+	+	February 2006 (johna)	version 2.1.4
+	+	added code to handle new pedantic version of gcc
+	+	made signed/unsigned char assignments explicit
+
+	+	February 21, 2007 - djonas	version 2.1.5
+	+	updated noip2.c: added SkipHeaders() instead of the magic 6 line pass
+	+	Changed to ip1.dynupdate.no-ip.com for ip retrieval
+
+	+	August 2007 (johna)	version 2.1.6
+	+	added fclose() for stdin, stdout & stderr to child
+	+	made Force_Update work on 30 day intervals
+
+	+	August 2007 (johna)	version 2.1.7
+	+	fixed bug introduced in 2.1.6 where errors from multiple
+	+	instances were not diplayed due to stderr being closed
+	+	added version number into shared mem and -S display
+
 */			
 /////////////////////////////////////////////////////////////////////////////                                            
 
@@ -116,6 +141,7 @@
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include <pwd.h>
+#include <time.h>
 
 #ifdef bsd_with_getifaddrs
 #include <ifaddrs.h>
@@ -141,11 +167,13 @@
 
 #define ENCRYPT			1
 
+#undef MAX
 #define MAX(x,y)		(((x)>(y))?(x):(y))
 
 #define READ_TIMEOUT		90
 #define WRITE_TIMEOUT		60
 #define CONNECT_TIMEOUT		60
+#define FORCE_INTERVAL		(1440 * 30)	// 30 days in minutes
 
 #define IPLEN			16
 #define LINELEN 	        256
@@ -153,14 +181,14 @@
 
 
 #define NOIP_NAME		"dynupdate.no-ip.com"
+#define NOIP_IP_SCRIPT	"ip1.dynupdate.no-ip.com"
 #define CLIENT_IP_PORT		8245
 
-#define VERSION			"2.1.1"
+#define VERSION			"2.1.7"
 #define USER_AGENT		"User-Agent: Linux-DUC/"VERSION
 #define SETTING_SCRIPT		"settings.php?"
 #define USTRNG			"username="
 #define PWDSTRNG		"&pass="
-#define GETIP_SCRIPT		"ip.php"
 #if ENCRYPT
   #define REQUEST		"requestL="
 #else
@@ -169,9 +197,9 @@
   #define UPDATE_SCRIPT		"ducupdate.php"
 
 #ifdef DEBUG
-  #define OPTCHARS		"CYU:Fc:dD:hp:u:x:SMi:K:I:"
+  #define OPTCHARS		"CYU:Fc:dD:hp:u:x:SMi:K:I:z"
 #else
-  #define OPTCHARS		"CYU:Fc:hp:u:x:SMi:K:I:"
+  #define OPTCHARS		"CYU:Fc:hp:u:x:SMi:K:I:z"
 #endif
 #define ARGC			1
 #define ARGF			(1<<1)
@@ -313,6 +341,9 @@ int	debug_toggle		=	0;
 int	kill_proc		=	0;
 int	reqnum			=	0;
 int	prompt_for_executable	=	1;
+int	shm_dump_active		=	0;
+int	Force_Update		=	0;	// force initial action
+int	dummy			=	0;
 void	*shmaddr		=	NULL;
 char	*program		=	NULL;
 char	*ourname		=	NULL;
@@ -387,7 +418,8 @@ struct NAMES {
 
 struct INSTANCE {
 	int	pid;
-	short	debug;
+	char	debug;
+	char	version;
 	short	interval;
 	char	Last_IP_Addr[IPLEN];
 	char	cfilename[LINELEN];
@@ -418,8 +450,11 @@ void	process_options(int argc, char *argv[]);
 void	alarm_handler(int signum);
 void	sigchld_handler(int signum);
 void	exit_handler(int signum);
-volatile void	wake_handler(int signum);
+void	wake_handler(int signum);
+char	put_version(char *ver);
+char	*get_version(unsigned char x);
 void	save_IP();
+void	dump_shm(struct shmid_ds *d);
 int	get_shm_info();
 void	get_our_visible_IPaddr(char *dest);
 void	getip(char *dest, char *device);
@@ -436,24 +471,25 @@ void	display_current_config();
 void	display_one_config(char *req, int interval, 
 			int nat, char *device, char *expath);
 int	parse_config();
-ushort	chksum(unsigned char *buf, unsigned int size);
+ushort	chksum(char *buf, unsigned int size);
 void	dump_buffer(int x);
 int	dynamic_update();
 int	handle_dynup_error(int error_code);
 int	handle_config_error(int err_code);
-void	*Malloc(unsigned int size);
+void	*Malloc(int size);
 int	get_xml_field(char *name, char *line, char *dest);
 void	add_to_list(char *gnm, struct NAMES *ns);
 int	yesno(char *fmt, ...);
 int	add_to_request(int kind, char *p);
 int	get_update_selection(int tgrp, int thst);
 int	GetNextLine(char *dest);
-void	url_encode(unsigned char *in, unsigned char *out);
+void SkipHeaders();
+void	url_encode(char *in, char *out);
 void	get_credentials(char *l, char *p);
 void	get_device_name(char *d);
 void	autoconf();
-int     bencode(const unsigned char *s, unsigned char *dst);
-int     bdecode(unsigned char *in, unsigned char *out);
+int     bencode(const char *s, char *dst);
+int     bdecode(char *in, char *out);
 void	Msg(char *fmt, ...);
 ///////////////////////////////////////////////////////////////////////////
 void Usage()
@@ -484,6 +520,7 @@ void Usage()
 	fprintf(stderr, "         -S               show configuration data\n");
 	fprintf(stderr, "         -M               permit multiple instances\n");
 	fprintf(stderr, "         -K processID     terminate instance PID\n");
+	fprintf(stderr, "         -z               activate shm dump code\n");
 	fprintf(stderr, "         -h               help (this text)\n");
 }
 ///////////////////////////////////////////////////////////////////////////
@@ -618,6 +655,7 @@ void process_options(int argc, char *argv[])
 		case 'd':
 			debug++;
 			log2syslog = -1;
+			shm_dump_active++;
 			strcat(saved_args, " -d");
 			break;
 		case 'D':
@@ -665,7 +703,9 @@ void process_options(int argc, char *argv[])
 			strcat(saved_args, optarg);
 			have_args |= ARGI;
 			break;
-
+		case 'z':
+			shm_dump_active++;
+			break;
 		default:
 			Usage();
 			exit(0);
@@ -725,8 +765,35 @@ void exit_handler(int signum)	//entered on SIGINT,  SIGTERM & SIGQUIT
 	// don't reset handler -- exit on second signal
 }
 ///////////////////////////////////////////////////////////////////////////
-volatile void wake_handler(int signum)	//entered on SIGUSR1
+void wake_handler(int signum)	//entered on SIGUSR1
 {			// used to wake sleeping process only
+	dummy++;
+}
+///////////////////////////////////////////////////////////////////////////
+char put_version(char *ver)
+{
+	char *p, x;
+	
+	p = ver + 2;	// step over 2.
+	x  = (*p -'0') * 10;	// get major num
+	p += 2;
+	x += (*p -'0');		// get minor num
+	return x;
+}
+///////////////////////////////////////////////////////////////////////////
+char *get_version(unsigned char z)
+{
+	int x, y;
+	static char vers[10];
+
+	if (z > 15) {
+	    x = z / 10;
+	    y = z % 10;
+	    sprintf(vers, "2.%d.%d", x, y);
+	} else {
+	    strcpy(vers, "unknown");
+	}
+	return vers;
 }
 ///////////////////////////////////////////////////////////////////////////
 void save_IP()
@@ -818,6 +885,26 @@ int config_file_inuse()
 	return retval;
 }
 /////////////////////////////////////////////////////////////////////////
+void dump_shm(struct shmid_ds *d)
+{
+	if (shm_dump_active) {
+	    fprintf(stderr, "\nShared memory info:\n");
+	    fprintf(stderr, "\tsize of struct shmid_ds is %d\n", (int)sizeof(struct shmid_ds));
+	    fprintf(stderr, "\tmode %o\n", d->shm_perm.mode & 0x1ff);
+	    fprintf(stderr, "\tuid %d, gid %d\n", 
+					d->shm_perm.uid, d->shm_perm.gid);
+	    fprintf(stderr, "\tcreated by uid %d, gid %d\n", 
+					d->shm_perm.cuid, d->shm_perm.cgid);
+	    fprintf(stderr, "\tsegment size = %d\n", (int)d->shm_segsz);
+	    fprintf(stderr, "\tlast attach time = %s",   ctime(&d->shm_atime));
+	    fprintf(stderr, "\tlast detach time = %s",   ctime(&d->shm_dtime));
+	    fprintf(stderr, "\tcreation time = %s",   ctime(&d->shm_ctime));
+	    fprintf(stderr, "\tnumber of attached procs = %d\n", (int)d->shm_nattch);
+	    fprintf(stderr, "\tcreation process id = %d\n", d->shm_cpid);
+	    fprintf(stderr, "\tlast user process id = %d\n", d->shm_lpid);
+	}
+}
+/////////////////////////////////////////////////////////////////////////
 int get_shm_info()
 {
 	int	i, flag;
@@ -836,14 +923,17 @@ int get_shm_info()
 	if ((ds.shm_nattch > 0) && (!multiple_instances)) { 
 	    Msg("One %s process is already active,", program);
 	    Msg("and the multiple instance flag (-M) is not set.");
+	    dump_shm(&ds);
 	    return FATALERR;
 	}
 	if (ds.shm_nattch >= MAX_INSTANCE) { 
 	    Msg("Too many %s processes active. Ending!", program);
+	    dump_shm(&ds);
 	    return FATALERR;
 	}
 	if ((shmaddr = shmat(shmid, 0, 0)) == (void *)-1) {
 	    Msg("Can't attach shared memory. (%s) Ending!", strerror(errno));
+	    dump_shm(&ds);
 	    return FATALERR;
 	}
 	shared = (struct SHARED *)shmaddr;
@@ -883,6 +973,7 @@ int get_shm_info()
 	my_instance->pid = getpid();
 	my_instance->debug = debug;
 	my_instance->interval = interval;
+	my_instance->version = put_version(VERSION);
 	sprintf(my_instance->Last_IP_Addr, new_config->lastIP);
 	sprintf(my_instance->args, "%s", saved_args);
 	sprintf(my_instance->cfilename, "%s", config_filename);
@@ -892,6 +983,7 @@ int get_shm_info()
 int run_as_background()
 {
 	int	x, delay;
+	int	time_to_update;
 	char	*err_string;
 
 	x = fork();
@@ -907,6 +999,9 @@ int run_as_background()
 		if (get_shm_info() == FATALERR)	
 		    return FATALERR;
 		log2syslog++;
+		fclose(stderr);
+		fclose(stdout);
+		fclose(stdin);
 		syslog(LOG_INFO, "v%s daemon started%s\n", 
 			VERSION, (nat) ?  " with NAT enabled" : "");
 		while (background) {
@@ -916,15 +1011,36 @@ int run_as_background()
 		    else
 			getip(IPaddress, device);
 #ifdef DEBUG
-		    if (my_instance->debug) 
+		    if (my_instance->debug)  {
 		        Msg("! Last_IP_Addr = %s, IP = %s",my_instance->Last_IP_Addr, IPaddress);
+			Msg("Force_Update == %d\n", Force_Update);
+		    }
 #endif
-		    if (*IPaddress && strcmp(IPaddress, my_instance->Last_IP_Addr))  {
+		    time_to_update = 0;
+		    if (*IPaddress) { 
+			if (strcmp(IPaddress, my_instance->Last_IP_Addr) ||
+				(Force_Update <= 0))
+			    time_to_update = 1;
+		    } else {	// make sure we don't lose force update trigger
+			if (Force_Update <= 0)
+			    Force_Update = my_instance->interval;
+		    }
+		    if (time_to_update) {
 			if (dynamic_update() == SUCCESS) 
 			    strcpy(my_instance->Last_IP_Addr, IPaddress);
-			*IPaddress = 0;
+			else		// don't lose forced update trigger
+			    Force_Update = MAX(Force_Update, 
+							my_instance->interval);
 			if (connect_fail)
-			    delay += 300;
+			    delay = MAX(300, delay); // wait at least 5 minutes more
+			*IPaddress = 0;
+		    }
+		    if (Force_Update <= 0) {
+			// reset to max 
+			Force_Update = FORCE_INTERVAL;
+		    } else {
+			// one time quanta closer to update trigger
+			Force_Update -= my_instance->interval;
 		    }
 		    if (background)	// signal may have reset this!
 			Sleep(delay);
@@ -1079,11 +1195,13 @@ char *despace(char *p)
 	return p;
 }
 ///////////////////////////////////////////////////////////////////////////
-ushort chksum(unsigned char *buf, unsigned int size)
+ushort chksum(char *p_buf, unsigned int size)
 {
 	int	i;
 	ushort	sum;
+	unsigned char *buf;
 
+	buf = (unsigned char *)p_buf;
 	sum = 0;
 	for (i=0; i<size; i++)
             sum += *buf++;
@@ -1268,7 +1386,8 @@ void display_current_config()
 			is = &shared->instance[i];
 			if (is->pid == 0)
 			    continue;
-			Msg("\nProcess %d, started as %s", is->pid, is->args);
+			Msg("\nProcess %d, started as %s, (version %s)", 
+				is->pid, is->args, get_version(is->version));
 			Msg("Using configuration from %s", is->cfilename);
 			Msg("Last IP Address set %s", is->Last_IP_Addr);
 			get_one_config(is->cfilename);
@@ -1372,7 +1491,7 @@ void get_our_visible_IPaddr(char *dest)
 	} else {
 	    // set up text for web page query
 	    sprintf(buffer, 
-		"GET http://%s/ip.php HTTP/1.0\r\n%s\r\n\r\n", NOIP_NAME,USER_AGENT);
+		"GET http://%s/ HTTP/1.0\r\n%s\r\n\r\n", NOIP_IP_SCRIPT,USER_AGENT);
 	    if ((x = converse_with_web_server()) < 0) {
 		handle_dynup_error(x);
 	    } else {
@@ -1387,7 +1506,7 @@ void get_our_visible_IPaddr(char *dest)
 		return;				// NORMAL EXIT
 	    } 
 	}
-	Msg("Can't get our visible IP address from %s", NOIP_NAME); 
+	Msg("Can't get our visible IP address from %s", NOIP_IP_SCRIPT); 
 	return;
 }
 ///////////////////////////////////////////////////////////////////////////
@@ -1405,7 +1524,7 @@ void dump_buffer(int x)
 int dynamic_update()
 {
 	int	i, x, is_group, retval, response;
-	char	*p, *pos, tbuf[LINELEN], gname[LINELEN];
+	char	*p, *pos, tbuf[LINELEN], ipbuf[LINELEN], gname[LINELEN];
 
 	retval = SUCCESS;
 	if ((x = Connect(port_to_use)) != SUCCESS) {
@@ -1431,8 +1550,7 @@ int dynamic_update()
 
 	// analyze results
 	offset = 0;
-	for (i=0; i<6; i++) 
-	    GetNextLine(tbuf);
+	SkipHeaders();
 	response = 0;
 	while (GetNextLine(tbuf)) {
 	    p = ourname = tbuf;	// ourname points at the host/group or error
@@ -1476,7 +1594,7 @@ int dynamic_update()
 		    Msg("Can't fork (%s) to run %s", strerror(errno), buffer);
 		    exit(1);
 		case 0:			// child
-		    sprintf(tbuf, "%s", IPaddress);
+		    sprintf(ipbuf, "%s", IPaddress);
 		    p = strrchr(buffer, '/');
 		    if (p)
 			p++;
@@ -1484,9 +1602,9 @@ int dynamic_update()
 			p = buffer;
 #ifdef DEBUG
 		    if (my_instance ? my_instance->debug : debug)
-			Msg("Running %s %s %s", buffer, p, tbuf);
+			Msg("Running %s %s %s %s", buffer, p, ipbuf, ourname);
 #endif
-	    	    execl(buffer, p, tbuf, ourname, (char *)0);
+	    	    execl(buffer, p, ipbuf, ourname, (char *)0);
 		    Msg("execl %s failed (%s)", buffer, strerror(errno));
 		    exit(0);		// bad execute -- don't care!
 		default:		// parent
@@ -1620,7 +1738,7 @@ int handle_config_error(int err_code)
 	return FATALERR;
 }
 //////////////////////////////////////////////////////////////////////////
-void *Malloc(unsigned int size)
+void *Malloc(int size)
 {
         void *x = malloc(size);
         if (x != NULL)
@@ -1870,6 +1988,19 @@ int get_update_selection(int tgrp, int thst)
 	return len;
 }
 //////////////////////////////////////////////////////////////////////////
+void SkipHeaders()
+{
+	char *p = &buffer[offset];
+	for(; *p; p++) {
+		if(strncmp(p, "\r\n\r\n", 4) == 0) {
+			offset += 4;
+			break;
+		}
+		offset++;
+	}
+	return;
+}
+//////////////////////////////////////////////////////////////////////////
 int GetNextLine(char *dest)
 {
         char    *p = &buffer[offset];
@@ -1894,9 +2025,12 @@ int GetNextLine(char *dest)
         return 0;
 }
 ///////////////////////////////////////////////////////////////////////////
-void url_encode(unsigned char *in, unsigned char *out)
+void url_encode(char *p_in, char *p_out)
 {
-        unsigned char ch;
+        unsigned char ch, *in, *out;
+
+	in  = (unsigned char *)p_in;
+	out = (unsigned char *)p_out;
 
         while ((ch = *in++)) {
             switch(ch) {
@@ -1906,13 +2040,13 @@ void url_encode(unsigned char *in, unsigned char *out)
 		case '@':  case '[': case '\\': case ']': case '^':
                 case '`': case '{': case '|': case '}': case '~':
 			*out++ = '%';
-			sprintf(out, "%2.2x", ch);
+			sprintf((char *)out, "%2.2x", ch);
 			out += 2;
 			break;
                 default: 
 		       if ((ch & 0x80) || (ch < 0x20)) {
                            *out++ = '%';
-                           sprintf(out, "%2.2x", ch);
+                           sprintf((char *)out, "%2.2x", ch);
                            out += 2;
                        } else
                            *out++ = ch;
@@ -1969,11 +2103,12 @@ void get_credentials(char *l, char *p)
 	url_encode(answer, p);
 }
 /////////////////////////////////////////////////////////////////////////////
-int get_all_device_names(unsigned char *devs)
+int get_all_device_names(char *devs)
 {
 	int		devnum=0;
 	unsigned char	*p, *q, *dq;
 
+
 #ifdef bsd_with_getifaddrs
 
 	struct ifaddrs *ifap, *ifa;
@@ -1981,7 +2116,7 @@ int get_all_device_names(unsigned char *devs)
 	    perror("getifaddrs()");
 	    exit(1);
 	}
-	dq = devs;     // point at name list
+	dq = (unsigned char *)devs;     // point at name list
 	for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
 		if (ifa->ifa_addr->sa_family == AF_LINK) {
 // FreeBSD doesn't define some of these!
@@ -2031,7 +2166,7 @@ int get_all_device_names(unsigned char *devs)
         }
 	ifrp = ibuf;
 	ifend = (struct ifreq *)((char *)ibuf + ifc.ifc_len);
-	dq = devs;     // point at name list
+	dq = (unsigned char *)devs;     // point at name list
 	for (; ifrp < ifend; ifrp = ifnext) {
 	    q = dq;     // add new name into list
 	    p = ifrp->ifr_name;
@@ -2072,13 +2207,13 @@ int get_all_device_names(unsigned char *devs)
 	    exit(1);
 	}
 	num_ifreq = Ifc.ifc_len / sizeof(struct ifreq);
-	dq = devs;     // add new name into list
+	dq = (unsigned char *)devs;     // add new name into list
 	for (pIfr=Ifc.ifc_req,i=0; i<num_ifreq; pIfr++,i++) {
 	    q = dq;     // add new name into list
-	    p = pIfr->ifr_name;
-	    if (strncmp("lo", p, 2) == 0)	// ignore loopbacks
+	    p = (unsigned char *)pIfr->ifr_name;
+	    if (strncmp("lo", (char *)p, 2) == 0)	// ignore loopbacks
 		continue;
-	    if (strchr(p, ':') != NULL)		// ignore 'virtual' devices
+	    if (strchr((char *)p, ':') != NULL)		// ignore 'virtual' devices
 		continue;
 	    devnum++;
 	    while (*p) 
@@ -2101,13 +2236,13 @@ void get_device_name(char *d)
 	int	i, devnum;
 	unsigned char devs[MAX_NET_DEVS][LINELEN];
 
-	devnum = get_all_device_names((unsigned char *)&devs);
+	devnum = get_all_device_names((char *)&devs);
 	switch (devnum) {
 	    case 0:
 		Msg(CMSG19);
 		exit(1);
 	    case 1: 		// most common case
-		strcpy(d, devs[0]);
+		strcpy(d, (char *)devs[0]);
 		break;
 	    default:		// firewall machine
 		Msg(CMSG20);
@@ -2119,7 +2254,7 @@ void get_device_name(char *d)
 		    fgets(answer, LINELEN, stdin);
 		    i = atoi(answer);
 		} while ((i < 0) || (i >= devnum));
-		strcpy(d, devs[i]);
+		strcpy(d, (char *)devs[i]);
 		break;
 	}
 }
@@ -2169,7 +2304,7 @@ void autoconf()
 	    Msg(CMSG27);
 	    return;
 	}
-        sprintf(buffer, "GET http://%s/%s\r\n\r\n", NOIP_NAME, GETIP_SCRIPT);
+        sprintf(buffer, "GET http://%s/\r\n\r\n", NOIP_IP_SCRIPT);
         if ((x = converse_with_web_server()) <= 0) {
             handle_dynup_error(x);
 	    return;
@@ -2211,8 +2346,7 @@ void autoconf()
         groups->glink = NULL;
         groups->nlink = NULL;
 
-	for (x=0; x<6; x++)	// skip first 6 lines HTTP protocol stuff
-            GetNextLine(line);
+	SkipHeaders();
         while (GetNextLine(line)) {
             p = line;
 	    s = settings;
@@ -2288,15 +2422,20 @@ void autoconf()
 	return;
 }
 /////////////////////////////////////////////////////////////////////////
-int bencode(const unsigned char *s, unsigned char *dst)
+int bencode(const char *p_s, char *p_dst)
 {                         // http basic authorization encoding (base64)
 #if ENCRYPT
-        int n           = strlen(s);
-        int n3byt       = n / 3; 
-        int k           = n3byt * 3; 
-        int nrest       = n % 3;
-        int i           = 0;
-        int dstlen      = 0;
+        int n, n3byt, k, i, nrest, dstlen;
+	unsigned char *s, *dst;
+
+	s  	= (unsigned char *)p_s;
+	dst	= (unsigned char *)p_dst;
+        n  	= strlen(p_s);
+        n3byt   = n / 3; 
+        k       = n3byt * 3; 
+        nrest   = n % 3;
+        i       = 0;
+        dstlen  = 0;
 
         while (i < k) {
           dst[dstlen++] = EncodeTable[(( s[i]  & 0xFC)>>2)];
@@ -2321,17 +2460,22 @@ int bencode(const unsigned char *s, unsigned char *dst)
         dst[dstlen] = 0;
 	return dstlen;
 #else
-        strcpy(dst, s);
-        return strlen(s);
+        strcpy(p_dst, p_s);
+        return strlen(p_s);
 
 #endif
 }
 ////////////////////////////////////////////////////////////////////////////
-int bdecode(unsigned char *in, unsigned char *out)
+int bdecode(char *p_in, char *p_out)
 {
 #if ENCRYPT
-        int x = strlen(in);
+	unsigned char *in, *out;
         unsigned char *p, *q, d1, d2, d3, d4;
+        int x;
+
+	in  = (unsigned char *)p_in;
+	out = (unsigned char *)p_out;
+        x = strlen(p_in);
 
         p = q = &in[x];
         while (x % B64MOD) {      // pad to a multiple of four (if malformed)
@@ -2355,9 +2499,8 @@ int bdecode(unsigned char *in, unsigned char *out)
         } while (x > 0);
         *out = 0;
 	*q = 0;					// replace original null
-        return 0;
 #else
-	strcpy(out, in);
+	strcpy(p_out, p_in);
 #endif
 	return 0;
 }
